;------------------------------------------------------------------------------

(define (initial-bindings node)
  (cond ((root? node)
         (error "Root nodes should not appear inside parse trees"))
        ((uns? node)
         '())
        ((cst? node)
         '())
        ((ref? node)
         '())
        ((set? node)
         (initial-bindings (set-val node)))
        ((cond? node)
         (append
           (initial-bindings (cond-pre node))
           (initial-bindings (cond-con node))
           (initial-bindings (cond-alt node))))
        ((proc? node)
         (append
           (map (lambda (var) (cons var set-empty)) (proc-parms node))
           (initial-bindings (proc-body node))))
        ((call? node)
         (apply append
           (initial-bindings (call-oper node))
           (map initial-bindings (call-args node))))
        (else
         (error "Unknown parse tree node type"))))

(define (print-values bindings)
  (for-each
    (lambda (x)
      (write (var-name (car x)))
      (display " =")
      (for-each
        (lambda (y)
          (display " ")
          (cond ((eq? y any-value) (write '*any*))
                ((proc? y) (write (parse-tree->expression y)))
                (else (write y))))
        (set->list (cdr x)))
      (newline))
    bindings))

(define any-value (cons 0 0))

(define (dflow-analysis ptree)
  (let ((bindings (initial-bindings ptree))
        (changes 0))
    (define (externally-reachable values)
      (for-each
        (lambda (x)
          (if (proc? x)
            (for-each
              (lambda (y)
                (update-bindings y (set-singleton any-value)))
              (proc-parms x))))
        values))
    (define (update-bindings var vals)
      (let ((location (assq var bindings)))
        (let ((old-values (cdr location)))
          (let ((new-values (set-union vals old-values)))
            (if (not (set-equal? new-values old-values))
              (begin
                (set-cdr! location new-values)
                (set! changes (+ changes 1))))))))
    (define (refine-values node)
      (cond ((root? node)
             (error "Root nodes should not appear inside parse trees"))
            ((uns? node)
             (set-singleton any-value))
            ((cst? node)
             (set-singleton (cst-val node)))
            ((ref? node)
             (if (var-bound (ref-var node))
               (cdr (assq (ref-var node) bindings))
               (set-singleton any-value)))
            ((set? node)
             (if (var-bound (set-var node))
               (update-bindings (set-var node) (refine-values (set-val node)))
               (externally-reachable (refine-values (set-val node))))
             (set-singleton any-value))
            ((cond? node)
             (let ((pre-values (refine-values (cond-pre node)))
                   (con-values (refine-values (cond-con node)))
                   (alt-values (refine-values (cond-alt node))))
               (set-union
                 (if (set-empty? (set-difference pre-values
                                   (set-singleton #f)))
                    set-empty
                    con-values)
                 (if (set-empty? (set-intersection pre-values
                                   (list->set (list #f any-value))))
                    set-empty
                    alt-values))))
            ((proc? node)
             (set-singleton node))
            ((call? node)
             (let ((oper-values (refine-values (call-oper node)))
                   (args-values (map refine-values (call-args node))))
               (apply set-union
                 (map (lambda (oper)
                        (cond ((eq? oper any-value)
                               (map externally-reachable args-values)
                               (set-singleton any-value))
                              ((proc? oper)
                               (for-each
                                 update-bindings
                                 (proc-parms oper)
                                 args-values)
                               (refine-values (proc-body oper)))
                              (else
                               set-empty)))
                      (set->list oper-values)))))
            (else
             (error "Unknown parse tree node type"))))
    (let loop ((ch changes))
      (externally-reachable (refine-values ptree))
      (if (= ch changes) bindings (loop changes)))))

;------------------------------------------------------------------------------
