(define (not x)
  (if (not x) #t #f))

(define (boolean? x)
  (if (eq? x #t) #t (eq? x #f)))

(define eqv? eqv?) ;--> (2)
(define eq? eq?) ;--> (2)

(define (equal? x y)
  (define (vector=? x y)
    (equal? (vector->list x) (vector->list y)))
  (cond ((eqv? x y)
         #t)
        ((pair? x)
         (if (pair? y)
           (if (equal? (car x) (car y)) (equal? (cdr x) (cdr y)) #f)
           #f))
        ((vector? x)
         (if (vector? y) (vector=? x y) #f))
        ((string? x)
         (if (string? y) (string=? x y) #f))
        (else
         #f)))

(define (pair? x)
  (if (pair? x) #t #f))

(define cons cons) ;--> (2)
(define car car) ;--> (1)
(define cdr cdr) ;--> (1)
(define set-car! set-car!) ;--> (2)
(define set-cdr! set-cdr!) ;--> (2)

(define (caar x) (car (car x)))
(define (cadr x) (car (cdr x)))
(define (cdar x) (cdr (car x)))
(define (cddr x) (cdr (cdr x)))
(define (caaar x) (car (caar x)))
(define (caadr x) (car (cadr x)))
(define (cadar x) (car (cdar x)))
(define (caddr x) (car (cddr x)))
(define (cdaar x) (cdr (caar x)))
(define (cdadr x) (cdr (cadr x)))
(define (cddar x) (cdr (cdar x)))
(define (cdddr x) (cdr (cddr x)))
(define (caaaar x) (car (caaar x)))
(define (caaadr x) (car (caadr x)))
(define (caadar x) (car (cadar x)))
(define (caaddr x) (car (caddr x)))
(define (cadaar x) (car (cdaar x)))
(define (cadadr x) (car (cdadr x)))
(define (caddar x) (car (cddar x)))
(define (cadddr x) (car (cdddr x)))
(define (cdaaar x) (cdr (caaar x)))
(define (cdaadr x) (cdr (caadr x)))
(define (cdadar x) (cdr (cadar x)))
(define (cdaddr x) (cdr (caddr x)))
(define (cddaar x) (cdr (cdaar x)))
(define (cddadr x) (cdr (cdadr x)))
(define (cdddar x) (cdr (cddar x)))
(define (cddddr x) (cdr (cdddr x)))

(define (null? x)
  (if (null? x) #t #f))

(define (list . l)
  l)

(define (length l)
  (let loop ((l l) (n 0))
    (if (pair? l) (loop (cdr l) (+ n 1)) n)))

(define (append l1 l2)
  (if (null? l1)
    l2
    (cons (car l1) (append (cdr l1) l2))))

(define (reverse l)
  (let loop ((l l) (x '()))
    (if (pair? l) (loop (cdr l) (cons (car l) x)) x)))

(define (list-tail l i)
  (if (= i 0) l (list-tail (cdr l) (- i 1))))

(define (list-ref l i)
  (car (list-tail l i)))

(define (last-pair l)
  (if (pair? (cdr l)) (last-pair (cdr l)) l))

(define (memq x l)
  (cond ((pair? l)
         (if (eq? (car l) x) l (memq x (cdr l))))
        (else
         #f)))

(define (memv x l)
  (cond ((pair? l)
         (if (eqv? (car l) x) l (memv x (cdr l))))
        (else
         #f)))

(define (member x l)
  (cond ((pair? l)
         (if (equal? (car l) x) l (member x (cdr l))))
        (else
         #f)))

(define (assq x l)
  (cond ((pair? l)
         (let ((couple (car l)))
           (if (eq? (car couple) x) couple (assq x (cdr l)))))
        (else
         #f)))

(define (assv x l)
  (cond ((pair? l)
         (let ((couple (car l)))
           (if (eqv? (car couple) x) couple (assv x (cdr l)))))
        (else
         #f)))

(define (assoc x l)
  (cond ((pair? l)
         (let ((couple (car l)))
           (if (equal? (car couple) x) couple (assoc x (cdr l)))))
        (else
         #f)))

(define (symbol? x)
  (if (symbol? x) #t #f))

(define symbol->string symbol->string) ;--> (1)
(define string->symbol string->symbol) ;--> (1)

(define number? number?) ;--> (1)
(define real? real?) ;--> (1)
(define integer? integer?) ;--> (1)

(define (odd? n)
  (not (even? n)))

(define (even? n)
  (= (remainder n 2) 0))

(define = =) ;--> (2)
(define < <) ;--> (2)
(define > >) ;--> (2)
(define <= <=) ;--> (2)
(define >= >=) ;--> (2)

(define (max x y)
  (if (> x y) x y))

(define (min x y)
  (if (< x y) x y))

(define + +) ;--> (0 1 2)
(define * *) ;--> (0 1 2)
(define - -) ;--> (1 2)
(define / /) ;--> (1 2)

(define (abs x)
  (if (< x 0) (- x) x))

(define (quotient x y)
  (/ x y))

(define (remainder x y)
  (- x (* (quotient x y) y)))

(define (expt n k)
  (if (= k 0)
    1
    (let ((x (expt (* n n) (quotient k 2))))
      (if (odd? k) (* n x) x))))

(define (number->string n)
  (define (loop k n i)
    (let ((x (quotient n 10)))
      (let ((s (if (= x 0) (make-string (+ i k)) (loop k x (+ i 1)))))
        (string-set! s
                     (- (string-length s) i)
                     (integer->char (+ 48 (- n (* x 10))))))))
  (if (< n 0) (string-set! (loop 1 (- n) 1) 0 #\-) (loop 0 n 1)))

(define (string->number s)
  (define (loop s i n)
    (let ((c (string-ref s i)))
      (and (char>=? c #\0)
           (char<=? c #\9)
           (let ((i (+ i 1)) (n (+ (* n 10) (- (char->integer c) 48))))
             (if (= (string-length s) i) n (loop s i n))))))
  (and (> (string-length s) 0)
       (cond ((char=? (string-ref s 0) #\-)
              (and (> (string-length s) 1)
                   (let ((n (loop s 1 0))) (and n (- n)))))
             ((char=? (string-ref s 0) #\+)
              (and (> (string-length s) 1)
                   (loop s 1 0)))
             (else
              (loop s 0 0)))))

(define (char? x)
  (if (char? x) #t #f))

(define char=? char=?) ;--> (2)
(define char<? char<?) ;--> (2)
(define char>? char>?) ;--> (2)
(define char<=? char<=?) ;--> (2)
(define char>=? char>=?) ;--> (2)

(define char->integer char->integer) ;--> (1)
(define integer->char integer->char) ;--> (1)

(define (char-uppercase c)
  (if (and (char>=? c #\a) (char<=? c #\z))
    (integer->char (- (char->integer c) 32))
    c))

(define (string? x)
  (if (string? x) #t #f))

(define make-string make-string) ;--> (1 2)
(define string-length string-length) ;--> (1)
(define string-ref string-ref) ;--> (2)
(define string-set! string-set!) ;--> (3)

(define (string=? x y)
  (if (= (string-length x) (string-length y))
    (let loop ((x x) (y y) (i (- (string-length x) 1)))
      (if (< i 0)
        #t
        (if (char=? (string-ref x i) (string-ref y i))
          (loop x y (- i 1))
          #f)))))

(define (string->list s)
  (let loop ((s s) (l '()) (i (- (string-length s) 1)))
    (if (< i 0)
      l
      (loop s (cons (string-ref s i) l) (- i 1)))))

(define (list->string l)
  (define (loop l i)
    (if (null? l)
      (make-string i)
      (string-set! (loop (cdr l) (+ i 1)) i (car l))))
  (loop l 0))

(define (vector? x)
  (if (vector? x) #t #f))

(define make-vector make-vector) ;--> (1 2)

(define (vector . l)
  (list->vector l))

(define vector-length vector-length) ;--> (1)
(define vector-ref vector-ref) ;--> (2)
(define vector-set! vector-set!) ;--> (3)

(define (vector->list v)
  (let loop ((v v) (l '()) (i (- (vector-length v) 1)))
    (if (< i 0)
      l
      (loop v (cons (vector-ref v i) l) (- i 1)))))

(define (list->vector l)
  (define (loop l i)
    (if (null? l)
      (make-vector i)
      (vector-set! (loop (cdr l) (+ i 1)) i (car l))))
  (loop l 0))

(define (procedure? x)
  (if (procedure? x) #t #f))

(define procedure-info procedure-info) ;--> (1)

(define apply apply) ;--> (2)

(define (map p l)
  (if (pair? l) (cons (p (car l)) (map p (cdr l))) '()))

(define (for-each p l)
  (if (pair? l) (begin (p (car l)) (for-each p (cdr l)))))

(define (make-delayed thunk)
  (cons #f thunk))

(define (force delayed)
  (if (not (car delayed))
    (begin
      (set-cdr! delayed ((cdr delayed)))
      (set-car! delayed #t)))
  (cdr delayed))

(define call-with-current-continuation call-with-current-continuation) ;--> (1)

(define (read)
  (define (read-non-whitespace-char)
    (let ((c (read-char)))
      (cond ((eof-object? c) c)
            ((char<=? c #\ ) (read-non-whitespace-char))
            (else            c))))
  (define (delimiter? c)
    (or (char<=? c #\ ) (memq c '(#\; #\( #\) #\# #\" #\' #\` #\,))))
  (define (read-list previous)
    (let ((c (read-non-whitespace-char)))
      (cond ((char=? c #\))
             (list previous))
            ((char=? c #\.)
             (let ((x (read)))
               (if (char=? (read-non-whitespace-char) #\))
                 (cons previous x)
                 (error "')' expected"))))
            (else
             (cons previous (read-list (rd c)))))))
  (define (read-vector)
    (define (loop i)
      (let ((c (read-non-whitespace-char)))
        (if (char=? c #\))
          (make-vector i)
          (let ((x (rd c))) (vector-set! (loop (+ i 1)) i x)))))
    (loop 0))
  (define (read-string)
    (define (loop i)
      (let ((c (read-char)))
        (cond ((char=? c #\")
               (make-string i))
              ((char=? c #\\)
               (let ((c (read-char))) (string-set! (loop (+ i 1)) i c)))
              (else
               (string-set! (loop (+ i 1)) i c)))))
    (loop 0))
  (define (read-symbol/number-string c)
    (define (loop i)
      (if (delimiter? (peek-char))
        (make-string i)
        (string-set! (loop (+ i 1)) i (char-uppercase (read-char)))))
    (string-set! (loop 1) 0 (char-uppercase c)))
  (define (rd c)
    (cond ((eof-object? c)
           c)
          ((char=? c #\()
           (if (char=? (peek-char) #\))
             (begin (read-char) '())
             (read-list (read))))
          ((char=? c #\#)
           (let ((c (char-uppercase (read-char))))
             (cond ((char=? c #\() (read-vector))
                   ((char=? c #\\) (read-char))
                   ((char=? c #\F) #f)
                   ((char=? c #\T) #t)
                   (else           (error "unknown '#' read macro" c)))))
          ((char=? c #\")
           (read-string))
          ((char=? c #\')
           (list 'quote (read)))
          ((char=? c #\`)
           (list 'quasiquote (read)))
          ((char=? c #\,)
           (if (char=? (peek-char) #\@)
             (begin (read-char) (list 'unquote-splicing (read)))
             (list 'unquote (read))))
          ((char=? c #\;)
           (let loop ()
             (if (not (char=? (read-char) (integer->char 10))) (loop)))
           (read))
          ((char=? c #\))
           (error "misplaced ')'"))
          (else
           (if (and (char=? c #\.) (delimiter? (peek-char)))
             (error "misplaced '.'"))
           (let ((s (read-symbol/number-string c)))
             (or (string->number s)
                 (string->symbol s))))))
  (rd (read-non-whitespace-char)))

(define eof-object? eof-object?) ;--> (1)
(define read-char read-char) ;--> (0)
(define peek-char peek-char) ;--> (0)

(define (write* obj display?)
  (define (write-string s display?)
    (if (not display?) (write-char #\"))
    (let loop ((n (string-length s)) (s s) (display? display?) (i 0))
      (if (< i n)
        (let ((c (string-ref s i)))
          (if (not display?) (if (char=? c #\") (write-char #\\)))
          (write-char c)
          (loop n s display? (+ i 1)))))
    (if (not display?) (write-char #\")))
  (define (write-chars s)
    (write-string s #t))
  (define (write-list l display?)
    (cond ((pair? l)
           (write-chars " ")
           (write* (car l) display?)
           (write-list (cdr l) display?))
          ((null? l)
           (write-chars ")"))
          (else
           (write-chars " . ")
           (write* l display?)
           (write-chars ")"))))
  (cond ((pair? obj)
         (write-chars "(")
         (write* (car obj) display?)
         (write-list (cdr obj) display?))
        ((null? obj)
         (write-chars "()"))
        ((eq? obj #f)
         (write-chars "#F"))
        ((eq? obj #t)
         (write-chars "#T"))
        ((symbol? obj)
         (write-chars (symbol->string obj)))
        ((number? obj)
         (write-chars (number->string obj)))
        ((string? obj)
         (write-string obj display?))
        ((vector? obj)
         (write-chars "#") (write* (vector->list obj) display?))
        ((procedure? obj)
         (let ((info (procedure-info obj)))
           (write-chars "#[procedure")
           (if info (begin (write-chars " ") (write* info #f)))
           (write-chars "]")))
        ((char? obj)
         (write-chars "#\\")
         (write-char obj))
        (else
         (write-chars "#[unknown]"))))

(define (write obj)
  (write* obj #f))

(define (display obj)
  (write* obj #t))

(define (newline)
  (write-char (integer->char 10)))

(define write-char write-char) ;--> (1)

(define (error msg . others)
  (display "error: ") (display msg) (newline)
  (let loop ((l others))
    (if (not (null? l))
      (begin
        (display "       ") (write (car l)) (newline)
        (loop (cdr l)))))
  (car 1))
