;------------------------------------------------------------------------------
;
; SET manipulation package
;
;------------------------------------------------------------------------------

(define (list->set list)    list)         ; convert list to set
(define (set->list set)     set)          ; convert set to list
(define set-empty           '())          ; the empty set
(define (set-empty? set)    (null? set))  ; is 'x' the empty set?
(define (set-member? x set) (memq x set)) ; is 'x' a member of the 'set'?
(define (set-singleton x)   (list x))     ; create a set containing only 'x'
(define (set-singleton? x)  (= (length x) 1)) ; is 'x' a singleton set?

(define (set-adjoin set x)                ; add the element 'x' to the 'set'
  (if (memq x set) set (cons x set)))

(define (set-remove set x)                ; remove the element 'x' from 'set'
  (cond ((null? set)       '())
        ((eq? (car set) x) (cdr set))
        (else              (cons (car set) (set-remove (cdr set) x)))))

(define (set-equal? s1 s2)
  (cond ((null? s1)         (null? s2))
        ((memq (car s1) s2) (set-equal? (cdr s1) (set-remove s2 (car s1))))
        (else               #f)))

(define (set-difference set . other-sets) ; return difference of sets
  (define (difference s1 s2)
    (cond ((null? s1)         '())
          ((memq (car s1) s2) (difference (cdr s1) s2))
          (else               (cons (car s1) (difference (cdr s1) s2)))))
  (n-ary difference set other-sets))

(define (set-union . sets)                ; return union of sets
  (define (union s1 s2)
    (cond ((null? s1)         s2)
          ((memq (car s1) s2) (union (cdr s1) s2))
          (else               (cons (car s1) (union (cdr s1) s2)))))
  (n-ary union '() sets))

(define (set-intersection set . other-sets) ; return intersection of sets
  (define (intersection s1 s2)
    (cond ((null? s1)         '())
          ((memq (car s1) s2) (cons (car s1) (intersection (cdr s1) s2)))
          (else               (intersection (cdr s1) s2))))
  (n-ary intersection set other-sets))

(define (n-ary function first rest)
  (if (null? rest)
    first
    (n-ary function (function first (car rest)) (cdr rest))))

(define (set-keep keep? set)
  (cond ((null? set)       '())
        ((keep? (car set)) (cons (car set) (set-keep keep? (cdr set))))
        (else              (set-keep keep? (cdr set)))))

(define (set-every? pred? set)
  (or (null? set)
      (and (pred? (car set))
           (set-every? pred? (cdr set)))))

(define (set-map proc set)
  (if (null? set)
    '()
    (cons (proc (car set)) (set-map proc (cdr set)))))

;------------------------------------------------------------------------------
