| -----------------------------------------------------------------------------
|
| This file contains definitions used by SIS's assembly language files.
|
| -----------------------------------------------------------------------------

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Constants:
|

|----- Scheme constants:

define(false,-1)		| constant #f
define(null,-3)			| constant ()
define(true,-5)			| constant #t
define(undefined,-5)		| constant for undefined values
define(forwarded_pair,-7)	| special value used by GC

|----- Type tags:

| for object-vectors:

vector_type	= 0x01	| type tag for vectors
symbol_type	= 0x02	| type tag for symbols

| for mixed-vectors:

stack_type	= 0x20	| type tag for stacks

| for byte-vectors:

string_type	= 0x30	| type tag for strings

| for procedures:

procedure_type	= 0x8000 | type tag for procedures with no closed variables
primitive_type	= 0xa000 | type tag for primitive procedures
closure_type	= 0xc000 | type tag for procedures with closed variables
entry_label_type= 0xe000 | type tag for entry labels

|----- Type checking masks:

fix_type_mask	= 0x01010101	| to test for fixnums
pair_type_mask	= 0x10101010	| to test for pairs
hunk_type_mask	= 0x44444444	| to test for hunks

|----- Mask to extract length from a hunk header:

length_mask	= 0x00fffffc

|----- Trap codes:

check_stack_trap	= 0
non_proc_trap		= 1
arity_trap		= 2
clo_arity_trap		= 3
wrong_type_trap		= 4
arith_ovfl_trap		= 5
out_bound_trap		= 6
not_impl_trap		= 9
cons_trap		= 10	| used when a cons overflows the heap
heap_trap		= 11	| used when the heap overflows

|----- Opcodes:

jump_opcode	= 0x4ef9			| 'jmp xxx.l' opcode
call_opcode	= 0x4eb9			| 'jsr xxx.l' opcode
trap_opcode	= 0x4e88+libr_ptr_reg		| 'jsr libr_ptr@' opcode
jump_prim_opcode= (0x4ee0+libr_ptr_reg)*65536	| 'jmp libr_ptr@(?)' opcode
call_prim_opcode= (0x4ea0+libr_ptr_reg)*65536	| 'jsr libr_ptr@(?)' opcode
lea_libr_opcode	= (0x41e0+libr_ptr_reg)*65536	| 'lea libr_ptr@(?),a?' opcode

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Register assignments:

define(DREG,`define($1,d$2)$1_reg = $2')
define(AREG,`define($1,a$2)$1_reg = $2+8')

nb_regs	= 3		| number of arguments that are passed in registers

DREG(arg_count,0)	| d0 contains the nb of arguments passed (16 bits)
DREG(first_arg,1)	| d1 contains the first argument
DREG(fix_mask,5)	| d5 contains the fixnum type mask (i.e. 0x01010101)
DREG(pair_mask,6)	| d6 contains the pair type mask (i.e. 0x10101010)
DREG(hunk_mask,7)	| d7 contains the hunk type mask (i.e. 0x44444444)

AREG(cont_reg,0)	| a0 contains the return address of the procedure
AREG(first_adr,1)	| a1 is the first temporary address register
AREG(empty_list,4)	| a4 contains the value () (i.e. the empty list)
AREG(libr_ptr,5)	| a5 points to the library procedures & trap handler
AREG(glob_ptr,6)	| a6 points to the global table
AREG(stack_ptr,7)	| a7 is the stack pointer

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| M4 definitions:
|

|----- Area manipulation:

define(AREA,`define(`THE_AREA',.$1).$1')


|----- Code deferral:

define(DEFERRED_CODE,`')
define(DEFER_CODE,`define(`DEFERRED_CODE',DEFERRED_CODE
$1:
$2)')

define(DUMP_DEFERRED_CODE,`define(`DEFERRED_CODE2',DEFERRED_CODE)define(`DEFERRED_CODE',`')DEFERRED_CODE2')

|----- Symbol generation:

define(THE_NUMBER,`0')
define(NEW_NUMBER,`define(`THE_NUMBER',incr(THE_NUMBER))THE_NUMBER')
define(NEW_SYMBOL,`define(`THE_SYMBOL',$1`'NEW_NUMBER)THE_SYMBOL')

|----- Conversion from Scheme symbols to assembler symbols:

define(SYM,`SYMLOOP(`$1',0)')
define(SYMLOOP,`ifelse($2,len(`$1'),,`SYMCHAR(substr(`$1',$2,1))`_'SYMLOOP(`$1',incr($2))')')
define(SYMCHAR,`ifelse($1,-,dash,$1,$,dollar,$1,?,question,$1,!,bang,$1,<,less,$1,=,equal,$1,>,greater,$1,+,plus,$1,*,star,$1,/,slash,$1)')

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Scheme object definition:
|

|----- to define pairs:

define(PAIR_OBJECT,`	.data1
	.long	$3
	.globl	$1
$1:
	.long	$2
	THE_AREA')

|----- to define symbols:

define(SYMBOL_OBJECT,`	.align	4
	.word	symbol_type*256
	.globl	$1
	.globl	$1
$1:
	.word	4
	.long	$2')

|----- to define strings:

define(STRING_OBJECT,`define(`OBJ_NAME',`$1')define(`END',`END_STRING')
	.align	4
	.word	string_type*256
	.globl	OBJ_NAME
OBJ_NAME:
	.word	OBJ_NAME.end-OBJ_NAME-2')

define(END_STRING,`OBJ_NAME.end:')

|----- to define vectors:

define(VECTOR_OBJECT,`define(`OBJ_NAME',`$1')define(`END',`END_VECTOR')
	.align	4
	.word	vector_type*256
	.globl	OBJ_NAME
OBJ_NAME:
	.word	OBJ_NAME.end-OBJ_NAME-2')

define(END_VECTOR,`OBJ_NAME.end:')

|----- to define procedures:

define(DEFINE_PRIMITIVE,`define(`PROC_STRING',`"`$1'"')define(`PROC_NAME',SYM(`$1'))define(`PROC_ARITY',$2)define(`END',`END_PROCEDURE')define(`LAST_CONST',0)
	.align	4
	.word	(LOCAL(end)-SCM.PROC_NAME.)/4+procedure_type
	.globl	SCM.PROC_NAME.
SCM.PROC_NAME.:
	ifelse($2,,,`ifelse($3,,`ENTER(`$2')',`ENTER_REST(`$2')')
	.globl	SCM.PROC_NAME.$2
SCM.PROC_NAME.$2:')')

define(DEFINE_PROCEDURE,`define(`PROC_STRING',`"`$1'"')define(`PROC_NAME',SYM(`$1'))define(`PROC_ARITY',$2)define(`END',`END_PROCEDURE')define(`LAST_CONST',0)
	.align	4
	.word	(LOCAL(end)-SCM.PROC_NAME.)/4+procedure_type
	.globl	SCM.PROC_NAME.
SCM.PROC_NAME.:
	ifelse($2,,,`ifelse($3,,`ENTER(`$2')',`ENTER_REST(`$2')')
	.globl	SCM.PROC_NAME.$2
SCM.PROC_NAME.$2:')')

define(LOCAL,`L.PROC_NAME.$1')

define(CONST,`define(`LAST_CONST',incr(LAST_CONST))LABEL(const`'LAST_CONST) .long $1')

define(END_PROCEDURE,`	.align	4
LABEL(end)
	.asciz	PROC_STRING
	.align	4')

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Virtual machine instructions:
|

|----- Recognizers:

define(NUM,`ifelse(index(translit(`$1',!-+0123456789,@!!!!!!!!!!!!),!),0,1,0)')
define(IS_TEMP,`ifelse(index(`$1',`temp('),0,1,0)')
define(IS_IMM,`ifelse(index(`$1',`imm('),0,1,0)')
define(IS_REG,`ifelse(index(`$1',`reg('),0,1,0)')
define(IS_STK,`ifelse(index(`$1',`stk('),0,1,0)')
define(IS_GLOB,`ifelse(index(`$1',`glob('),0,1,0)')
define(IS_INDX,`ifelse(index(`$1',`indx('),0,1,0)')
define(IS_CONST,`ifelse(index(`$1',`const('),0,1,0)')
define(IS_LBL,`ifelse(index(`$1',`lbl('),0,1,0)')
define(IS_PRIM,`ifelse(index(`$1',`prim('),0,1,0)')

define(DATA_REG,`ifelse(`$1',d0,1,`$1',d1,1,`$1',d2,1,`$1',d3,1,`$1',d4,1,`$1',d5,1,`$1',d6,1,`$1',d7,1,0)')
define(ADDR_REG,`ifelse(`$1',a0,1,`$1',a1,1,`$1',a2,1,`$1',a3,1,`$1',a4,1,`$1',a5,1,`$1',a6,1,`$1',a7,1,0)')
define(IMM,`ifelse(index(`$1',`#'),0,1,0)')

define(GET_INTO_DATA_REG,`ifelse(DATA_REG(`$1'),1,`define(`THE_DATA_REG',$1)',`$2',,`movl	$1,d0
	define(`THE_DATA_REG',d0)',`GET_INTO_DATA_REG($2,$3,$4,$5,$6,$7,$8,$9)')')

define(GET_INTO_ADDR_REG,`ifelse(ADDR_REG(`$1'),1,`define(`THE_ADDR_REG',$1)',`$2',,`movl	$1,a1
	define(`THE_ADDR_REG',a1)',`GET_INTO_ADDR_REG($2,$3,$4,$5,$6,$7,$8,$9)')')

|----- OPERAND conversion:

define(OPERAND,`ifelse(`$2',,`OPND(`$1',a1)',`OPND(`$1',`$2')')')

define(OPND,`ifelse(IS_INDX(`$1'),1,`OPERAND(PARAM(`$1',1),`$2')ifelse(ADDR_REG(THE_OPND),0,`movl	THE_OPND,$2
	define(`THE_OPND',$2)')define(`THE_OPND',THE_OPND@ifelse(PARAM(`$1',2),0,,(PARAM(`$1',2))))',`define(`THE_OPND',NON_INDX_OPND(`$1'))')')

define(NON_INDX_OPND,`ifelse(IS_REG(`$1'),1,`ifelse(PARAM(`$1',1),0,`cont_reg',`d`'PARAM(`$1',1)')',IS_STK(`$1'),1,`ifelse(PARAM(`$1',1),1,`stack_ptr@',`stack_ptr@(PARAM(`$1',1)*4-4)')',IS_GLOB(`$1'),1,`glob_ptr@(PARAM(`$1',1)*6+256)',IS_IMM(`$1'),1,`ifelse(PARAM(`$1',1),null,`empty_list',`#PARAM(`$1',1)')',IS_CONST(`$1'),1,`LOCAL(const`'PARAM(`$1',1))',IS_TEMP(`$1'),1,`a`'PARAM(`$1',1)',`$1',pop,`stack_ptr@+',`$1',push,`stack_ptr@-',`$1')')

define(PARAM,`PARAM2($2,substr(`$1',eval(index(`$1',`(')+1))')
define(PARAM2,`ifelse($1,1,`$2',`PARAM2(eval($1-1),`$3',`$4',`$5',`$6',`$7',`$8',`$9')')')

|----- COPY instruction:

define(COPY,`ifelse(IS_GLOB(`$2'),1,`COPY_GLOB(`$1',PARAM(`$2',1))',`COPY_NORM(`$1',`$2')')')
define(COPY_GLOB,`ifelse(`$1',reg(1),`COPY_GLOB2(`$2')',`movl	d1,d0
	COPY_NORM(`$1',reg(1))
	COPY_GLOB2(`$2')
	movl	d0,d1')')
define(COPY_GLOB2,`lea	glob_ptr@($1*6+256),a1
	jsr	glob_ptr@(eval((`$1'-1)*3/4%32*8+4))')
define(COPY_NORM,`ifelse(IS_PRIM(`$1'),1,`OPERAND(`$2')lea	SCM.SYM(PARAM(`$1',1)).PARAM(`$1',2),ifelse(ADDR_REG(THE_OPND),1,THE_OPND,a2)`'ifelse(ADDR_REG(THE_OPND),1,,`
	movl	a2,THE_OPND')',IS_LBL(`$1'),1,`OPERAND(`$2')ifelse(ADDR_REG(THE_OPND),1,`lea	LOCAL(PARAM(`$1',1)),THE_OPND',`lea	LOCAL(PARAM(`$1',1)),a2
	movl	a2,THE_OPND')',`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)ifelse(IMM(OPND1),1,`ifelse(DATA_REG(THE_OPND),1,`ifelse(SMALL_NUM(PARAM(`$1',1)),1,`moveq	OPND1,THE_OPND',`movl	OPND1,THE_OPND')',`movl	OPND1,THE_OPND')',`movl	OPND1,THE_OPND')')')

define(SMALL_NUM,`ifelse(eval(`$1'<=127),1,eval(`$1'>=-128),0)')

|----- DEALLOC instruction:

define(DEALLOC,`ifelse(`$1',0,,eval(`$1'<=2),1,`addql	#$1*4,stack_ptr
	',eval(`$1'<=8191),1,`addw	#$1*4,stack_ptr
	',`addl	#$1*4,stack_ptr
	')')

|----- LABEL instruction:

define(LABEL,`LOCAL($1):')

|----- ENTRY_LABEL instruction:

define(ENTRY_LABEL,
`	.align	4
	.word	(LOCAL($1)-SCM.PROC_NAME.)/4+entry_label_type
LABEL($1)')

|----- conditional branch instructions:

define(IF,`OPERAND(`$1')cmpl	THE_OPND,empty_list
	bhi	LOCAL($2)')

define(IF_NOT,`OPERAND(`$1')cmpl	THE_OPND,empty_list
	bls	LOCAL($2)')

define(IF_NULL,`OPERAND(`$1')cmpl	THE_OPND,empty_list
	beq	LOCAL($2)')

define(IF_NOT_NULL,`OPERAND(`$1')cmpl	THE_OPND,empty_list
	bne	LOCAL($2)')

define(IF_FIXNUM,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,fix_mask
	bne	LOCAL($2)')

define(IF_NOT_FIXNUM,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,fix_mask
	beq	LOCAL($2)')

define(IF_FLONUM,`define(`IF_FLONUM_LABEL',NEW_SYMBOL(z))OPERAND(`$1')movl	THE_OPND,d0
	btst	#0,d0
	beqs	LOCAL(IF_FLONUM_LABEL)
	swap	d0
	cmpw	#-2,d0
	bcs	LOCAL($2)
LABEL(IF_FLONUM_LABEL)')

define(IF_NOT_FLONUM,`OPERAND(`$1')movl	THE_OPND,d0
	btst	#0,d0
	beq	LOCAL($2)
	swap	d0
	cmpw	#-2,d0
	bhs	LOCAL($2)')

define(IF_CHAR,`define(`IF_CHAR_LABEL',NEW_SYMBOL(z))OPERAND(`$1')movl	THE_OPND,d0
	btst	#0,d0
	beqs	LOCAL(IF_CHAR_LABEL)
	swap	d0
	cmpw	#-2,d0
	beq	LOCAL($2)
LABEL(IF_CHAR_LABEL)')

define(IF_NOT_CHAR,`OPERAND(`$1')movl	THE_OPND,d0
	btst	#0,d0
	beq	LOCAL($2)
	swap	d0
	cmpw	#-2,d0
	bne	LOCAL($2)')

define(IF_PAIR,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,pair_mask
	bne	LOCAL($2)')

define(IF_NOT_PAIR,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,pair_mask
	beq	LOCAL($2)')

define(IF_HUNK,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	bne	LOCAL($2)')

define(IF_NOT_HUNK,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beq	LOCAL($2)')

define(IF_PROC,`define(`IF_PROC_LABEL',NEW_SYMBOL(z))OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beqs	LOCAL(IF_PROC_LABEL)
	GET_INTO_ADDR_REG(THE_OPND,THE_DATA_REG)tstw	THE_ADDR_REG@(-2)
	bmi	LOCAL($2)
LABEL(IF_PROC_LABEL)')

define(IF_NOT_PROC,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beq	LOCAL($2)
	GET_INTO_ADDR_REG(THE_OPND,THE_DATA_REG)tstw	THE_ADDR_REG@(-2)
	bpl	LOCAL($2)')

define(IF_STRING,`define(`IF_STRING_LABEL',NEW_SYMBOL(z))OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beqs	LOCAL(IF_STRING_LABEL)
	GET_INTO_ADDR_REG(THE_OPND,THE_DATA_REG)cmpb	#string_type,THE_ADDR_REG@(-2)
	beq	LOCAL($2)
LABEL(IF_STRING_LABEL)')

define(IF_NOT_STRING,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beq	LOCAL($2)
	GET_INTO_ADDR_REG(THE_OPND,THE_DATA_REG)cmpb	#string_type,THE_ADDR_REG@(-2)
	bne	LOCAL($2)')

define(IF_VECTOR,`define(`IF_VECTOR_LABEL',NEW_SYMBOL(z))OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beqs	LOCAL(IF_VECTOR_LABEL)
	GET_INTO_ADDR_REG(THE_OPND,THE_DATA_REG)cmpb	#vector_type,THE_ADDR_REG@(-2)
	beq	LOCAL($2)
LABEL(IF_VECTOR_LABEL)')

define(IF_NOT_VECTOR,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beq	LOCAL($2)
	GET_INTO_ADDR_REG(THE_OPND,THE_DATA_REG)cmpb	#vector_type,THE_ADDR_REG@(-2)
	bne	LOCAL($2)')

define(IF_SYMBOL,`define(`IF_SYMBOL_LABEL',NEW_SYMBOL(z))OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beqs	LOCAL(IF_SYMBOL_LABEL)
	GET_INTO_ADDR_REG(THE_OPND,THE_DATA_REG)cmpb	#symbol_type,THE_ADDR_REG@(-2)
	beq	LOCAL($2)
LABEL(IF_SYMBOL_LABEL)')

define(IF_NOT_SYMBOL,`OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beq	LOCAL($2)
	GET_INTO_ADDR_REG(THE_OPND,THE_DATA_REG)cmpb	#symbol_type,THE_ADDR_REG@(-2)
	bne	LOCAL($2)')

define(IF_EQ,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)ifelse(ADDR_REG(OPND1),1,`cmpl	THE_OPND,OPND1',ADDR_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND',DATA_REG(OPND1),1,`cmpl	THE_OPND,OPND1',DATA_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND',IMM(OPND1),1,`cmpl	OPND1,THE_OPND',IMM(THE_OPND),1,`cmpl	THE_OPND,OPND1',`GET_INTO_DATA_REG(OPND1)cmpl	THE_OPND,THE_DATA_REG')
	beq	LOCAL($3)')

define(IF_NOT_EQ,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)ifelse(ADDR_REG(OPND1),1,`cmpl	THE_OPND,OPND1',ADDR_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND',DATA_REG(OPND1),1,`cmpl	THE_OPND,OPND1',DATA_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND',IMM(OPND1),1,`cmpl	OPND1,THE_OPND',IMM(THE_OPND),1,`cmpl	THE_OPND,OPND1',`GET_INTO_DATA_REG(OPND1)cmpl	THE_OPND,THE_DATA_REG')
	bne	LOCAL($3)')

define(IF_LT,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)ifelse(ADDR_REG(OPND1),1,`cmpl	THE_OPND,OPND1 
	blt	LOCAL(`$3')',ADDR_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND
	bgt	LOCAL(`$3')',DATA_REG(OPND1),1,`cmpl	THE_OPND,OPND1 
	blt	LOCAL(`$3')',DATA_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND 
	bgt	LOCAL(`$3')',IMM(OPND1),1,`cmpl	OPND1,THE_OPND 
	bgt	LOCAL(`$3')',IMM(THE_OPND),1,`cmpl	THE_OPND,OPND1 
	blt	LOCAL(`$3')',`GET_INTO_DATA_REG(OPND1)cmpl	THE_OPND,THE_DATA_REG
	blt	LOCAL(`$3')')')

define(IF_NOT_LT,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)ifelse(ADDR_REG(OPND1),1,`cmpl	THE_OPND,OPND1 
	bge	LOCAL(`$3')',ADDR_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND
	ble	LOCAL(`$3')',DATA_REG(OPND1),1,`cmpl	THE_OPND,OPND1 
	bge	LOCAL(`$3')',DATA_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND 
	ble	LOCAL(`$3')',IMM(OPND1),1,`cmpl	OPND1,THE_OPND 
	ble	LOCAL(`$3')',IMM(THE_OPND),1,`cmpl	THE_OPND,OPND1 
	bge	LOCAL(`$3')',`GET_INTO_DATA_REG(OPND1)cmpl	THE_OPND,THE_DATA_REG
	bge	LOCAL(`$3')')')

define(IF_GT,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)ifelse(ADDR_REG(OPND1),1,`cmpl	THE_OPND,OPND1 
	bgt	LOCAL(`$3')',ADDR_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND
	blt	LOCAL(`$3')',DATA_REG(OPND1),1,`cmpl	THE_OPND,OPND1 
	bgt	LOCAL(`$3')',DATA_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND 
	blt	LOCAL(`$3')',IMM(OPND1),1,`cmpl	OPND1,THE_OPND 
	blt	LOCAL(`$3')',IMM(THE_OPND),1,`cmpl	THE_OPND,OPND1 
	bgt	LOCAL(`$3')',`GET_INTO_DATA_REG(OPND1)cmpl	THE_OPND,THE_DATA_REG
	bgt	LOCAL(`$3')')')

define(IF_NOT_GT,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)ifelse(ADDR_REG(OPND1),1,`cmpl	THE_OPND,OPND1 
	ble	LOCAL(`$3')',ADDR_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND
	bge	LOCAL(`$3')',DATA_REG(OPND1),1,`cmpl	THE_OPND,OPND1 
	ble	LOCAL(`$3')',DATA_REG(THE_OPND),1,`cmpl	OPND1,THE_OPND 
	bge	LOCAL(`$3')',IMM(OPND1),1,`cmpl	OPND1,THE_OPND 
	bge	LOCAL(`$3')',IMM(THE_OPND),1,`cmpl	THE_OPND,OPND1 
	ble	LOCAL(`$3')',`GET_INTO_DATA_REG(OPND1)cmpl	THE_OPND,THE_DATA_REG
	ble	LOCAL(`$3')')')

|----- JUMP and CALL instructions:

define(JUMP,`ifelse(IS_LBL(`$1'),1,`JUMP_LABEL(PARAM(`$1',1),`$2',`$3')',IS_PRIM(`$1'),1,`JUMP_PRIM(SCM.SYM(PARAM(`$1',1)).PARAM(`$1',2),`$2',`$3')',POP_JUMP(`$1',`$2'),1,`SETUP_ARG_COUNT(`$3')rts',IS_GLOB(`$1'),1,`JUMP_GLOB(PARAM(`$1',1),`$2',`$3')',`JUMP_NORM(`$1',`$2',`$3')')`'DUMP_DEFERRED_CODE')
define(JUMP_CHECK,`ifelse(IS_LBL(`$1'),1,`JUMP_LABEL(PARAM(`$1',1),`$2',`$3')',IS_PRIM(`$1'),1,`JUMP_PRIM(SCM.SYM(PARAM(`$1',1)).PARAM(`$1',2),`$2',`$3')',IS_GLOB(`$1'),1,`JUMP_GLOB(PARAM(`$1',1),`$2',`$3')',`JUMP_CHECK_NORM(`$1',`$2',`$3')')`'DUMP_DEFERRED_CODE')

define(JUMP_LABEL,`DEALLOC(`$2')SETUP_ARG_COUNT(`$3')bra	LOCAL(`$1')')
define(JUMP_PRIM,`DEALLOC(`$2')SETUP_ARG_COUNT(`$3')bra	$1')
define(JUMP_GLOB,`DEALLOC(`$2')SETUP_ARG_COUNT(`$3')jmp	glob_ptr@($1*6+254)')
define(JUMP_NORM,`OPERAND(`$1')GET_INTO_ADDR_REG(THE_OPND)DEALLOC(`$2')SETUP_ARG_COUNT(`$3')jmp	THE_ADDR_REG@')
define(JUMP_CHECK_NORM,`define(`JUMP_CHECK_NORM_LABEL',NEW_SYMBOL(e))OPERAND(`$1')GET_INTO_DATA_REG(THE_OPND)btst	THE_DATA_REG,hunk_mask
	beqs	LOCAL(JUMP_CHECK_NORM_LABEL)
	GET_INTO_ADDR_REG(THE_OPND,THE_DATA_REG)tstw	THE_ADDR_REG@(-2)
	bpl	LOCAL(JUMP_CHECK_NORM_LABEL)
	DEALLOC(`$2')SETUP_ARG_COUNT(`$3')jmp	THE_ADDR_REG@
LABEL(JUMP_CHECK_NORM_LABEL)
	ERROR(non_proc_trap)')

define(CALL,`define(`CALL_LABEL',NEW_SYMBOL(c))COPY(lbl(CALL_LABEL),reg(0))
	JUMP(`$1',`$2',`$3')
ENTRY_LABEL(CALL_LABEL)')

define(CALL_CHECK,`define(`CALL_CHECK_LABEL',NEW_SYMBOL(c))COPY(lbl(CALL_CHECK_LABEL),reg(0))
	JUMP(`$1',`$2',`$3')
ENTRY_LABEL(CALL_CHECK_LABEL)')

define(POP_JUMP,`ifelse($1,pop,`ifelse($2,0,1,0)',0)')

define(SETUP_ARG_COUNT,`ifelse($1,,,$1,0,`moveq	#1,arg_count	| no arg
	',$1,1,`moveq	#-1,arg_count	| 1 arg
	',$1,2,`moveq	#0,arg_count	| 2 args
	',SMALL_NUM(eval($1+1)),1,`moveq	#eval($1+1),arg_count	| $1 args
	',`movw	#eval($1+1),arg_count	| $1 args
	')')

|----- OPEN_CODE instruction:

define(OPEN_CODE,`OPEN_CODE2(SYM(PARAM(`$1',1)),PARAM(`$1',2),`$2',`$3',`$4',`$5',`$6',`$7',`$8')')
define(OPEN_CODE2,`$1_$2_OPEN_CODE(`$3',`$4',`$5',`$6',`$7',`$8',`$9')')

define(SYM(`+')_2_OPEN_CODE,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)define(`OPND2',THE_OPND)OPERAND(`$3',a3)ifelse(OPND1,THE_OPND,`ADD2(OPND2,THE_OPND)',OPND2,THE_OPND,`ADD2(OPND1,THE_OPND)',`movl	OPND1,THE_OPND
	ADD2(OPND2,THE_OPND)')')

define(ADD2,`ifelse(`$1',#8,`addql	#8,$2',`$1',#-8,`subql	#8,$2',`addl	$1,$2')')

define(SYM(`-')_2_OPEN_CODE,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)define(`OPND2',THE_OPND)OPERAND(`$3',a3)ifelse(OPND1,THE_OPND,`SUB2(OPND2,THE_OPND)',OPND2,THE_OPND,`movl	OPND1,d0
	SUB2(OPND2,d0)
	movl	d0,THE_OPND',`movl	OPND1,THE_OPND
	SUB2(OPND2,THE_OPND)')')

define(SUB2,`ifelse(`$1',#8,`subql	#8,$2',`$1',#-8,`addql	#8,$2',`subl	$1,$2')')

define(SYM(`*')_2_OPEN_CODE,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)define(`OPND2',THE_OPND)OPERAND(`$3',a3)ifelse(OPND1,THE_OPND,`MUL2(OPND2,THE_OPND)',OPND2,THE_OPND,`MUL2(OPND1,THE_OPND)',`movl	OPND1,THE_OPND
	MUL2(OPND2,THE_OPND)')')

define(MUL2,`GET_INTO_DATA_REG(`$2')asrl	#3,THE_DATA_REG
	mulsl	$1,THE_DATA_REG
	movl	THE_DATA_REG,$2')

define(SYM(`/')_2_OPEN_CODE,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)define(`OPND2',THE_OPND)OPERAND(`$3',a3)ifelse(OPND1,THE_OPND,`DIV2(OPND2,THE_OPND)',OPND2,THE_OPND,`movl	OPND1,d0
	DIV2(OPND2,d0)
	movl	d0,THE_OPND',`movl	OPND1,THE_OPND
	DIV2(OPND2,THE_OPND)')')

define(DIV2,`GET_INTO_DATA_REG(`$2')divsl	$1,THE_DATA_REG
	asll	#3,THE_DATA_REG
	movl	THE_DATA_REG,$2')

define(SYM(`CAR')_1_OPEN_CODE,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)GET_INTO_ADDR_REG(OPND1)movl	THE_ADDR_REG@,THE_OPND')

define(SYM(`CDR')_1_OPEN_CODE,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)GET_INTO_ADDR_REG(OPND1)movl	THE_ADDR_REG@-,THE_OPND')

define(SYM(`SET-CAR!')_2_OPEN_CODE,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)define(`OPND2',THE_OPND)OPERAND(`$3',a3)GET_INTO_ADDR_REG(OPND1)movl	OPND2,THE_ADDR_REG@
	movl	THE_ADDR_REG,THE_OPND')

define(SYM(`SET-CDR!')_2_OPEN_CODE,`OPERAND(`$1')define(`OPND1',THE_OPND)OPERAND(`$2',a2)define(`OPND2',THE_OPND)OPERAND(`$3',a3)GET_INTO_ADDR_REG(OPND1)movl	OPND2,THE_ADDR_REG@(-4)
	movl	THE_ADDR_REG,THE_OPND')

|----- CHECK_STACK instruction:

define(CHECK_STACK,`define(`CHECK_STACK_LABEL1',NEW_SYMBOL(t))define(`CHECK_STACK_LABEL2',NEW_SYMBOL(r))cmpl	glob_ptr@,stack_ptr
	bcs	LOCAL(CHECK_STACK_LABEL1)`'DEFER_CODE(LOCAL(CHECK_STACK_LABEL1),`	ERROR(check_stack_trap)
	bra	LOCAL(CHECK_STACK_LABEL2)')
LABEL(CHECK_STACK_LABEL2)')

|----- MAKE_CELL instruction:

define(MAKE_CELL,`define(`MAKE_CELL_LABEL',NEW_SYMBOL(c))COPY(lbl(MAKE_CELL_LABEL),temp(1))
	OPERAND(`$1',a2)movl	THE_OPND,a2
	bra	make_cell
ENTRY_LABEL(MAKE_CELL_LABEL)
	OPERAND(`$1',a1)movl	a3,THE_OPND')

|----- MAKE_CLOSURES instruction:

define(MAKE_CLOSURES,`define(`MAKE_CLOSURES_LABEL',NEW_SYMBOL(c))COPY(lbl(MAKE_CLOSURES_LABEL),temp(1))
	moveq	#CLOSURE_TOTAL_SIZE(`$1'),d0
	bra	make_closures
ENTRY_LABEL(MAKE_CLOSURES_LABEL)
CLOSURE_PUSH(`$1',0)')

define(CLOSURE_TOTAL_SIZE,`eval(2+ifelse(index(`$1',;),-1,`$1',`substr(`$1',0,index(`$1',;))+CLOSURE_TOTAL_SIZE(substr(`$1',eval(index(`$1',;)+1)))'))')

define(CLOSURE_PUSH,`	movl	#(ifelse(index(`$1',;),-1,`$1',`substr(`$1',0,index(`$1',;))')+1+closure_type)*65536+call_opcode,ifelse(`$2',0,`a2@
	pea	a2@(2)',`a2@($2*4)
	pea	a2@($2*4+2)')
ifelse(index(`$1',;),-1,,`CLOSURE_PUSH(substr(`$1',eval(index(`$1',;)+1)),eval(`$2'+substr(`$1',0,index(`$1',;))+2))')')

|----- ENTER instructions:

define(ENTER,`define(`ENTER_LABEL',NEW_SYMBOL(t))DEFER_CODE(LOCAL(ENTER_LABEL),`	ERROR(arity_trap)')ifelse(`$1',1,`bpl	LOCAL(ENTER_LABEL)',`$1',2,`bne	LOCAL(ENTER_LABEL)',`ifelse(eval(`$1'<8),1,subqw,subw)	#$1+1,arg_count
	bne	LOCAL(ENTER_LABEL)')')

define(ENTER_CLO,`define(`ENTER_LABEL',NEW_SYMBOL(t))DEFER_CODE(LOCAL(ENTER_LABEL),`	ERROR(clo_arity_trap)')ifelse(`$1',1,`bpl	LOCAL(ENTER_LABEL)',`$1',2,`bne	LOCAL(ENTER_LABEL)',`ifelse(eval(`$1'<8),1,subqw,subw)	#$1+1,arg_count
	bne	LOCAL(ENTER_LABEL)')
	subql	#6,stack_ptr@')

define(ENTER_REST,`bsr	enter_rest
	.word	$1-1')

define(ENTER_REST_CLO,`bsr	enter_rest_clo
	.word	$1-1')

define(IF_NOT_0ARG,`subqw	#1,d0
	bnes	LOCAL($1)
	.globl	SCM.PROC_NAME.0
SCM.PROC_NAME.0:')

define(IF_NOT_1ARG,`bpls	LOCAL($1)
	.globl	SCM.PROC_NAME.1
SCM.PROC_NAME.1:')

define(IF_NOT_2ARGS,`bnes	LOCAL($1)
	.globl	SCM.PROC_NAME.2
SCM.PROC_NAME.2:')

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

|----- trap handler calling

define(TRAP,`jsr	libr_ptr@
	.word	(.-SCM.PROC_NAME.)/4*16+$1')

|----- error handler calling

define(ERROR,`TRAP($1)')

|----- heap overflow on cons:

define(TRAP_CONS,`TRAP(cons_trap)')

|----- heap overflow:

define(TRAP_HEAP,`TRAP(heap_trap)')

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	AREA(data2)
