|------------------------------------------------------------------------------
|
| This file contains the garbage collector used by SIS.
|
|------------------------------------------------------------------------------


| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Representation of objects:
|
| Objects are represented using 32 bit values.  When more than 32 bits
| are needed to represent an object, the 32 bits are actually a pointer
| to the object in memory.
|
|                                28     28
| Fixnum (integer in the range -2   .. 2  -1):
|
|     XXXXXXXXXXXXXXXXXXXXXXXXXXXXX000
|     \------ integer value ------/
|
| Floating point number and other constants:
|
|     XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX1
|     \-- encoding of f.p. number --/
|              or constant
|   for example:
|     11111111111111111111111111111111 = #f
|     11111111111111111111111111111101 = ()
|     11111111111111111111111111111011 = #t
|     11111111111111111111111111111001 = undefined value
|     1111111111111110XXXXXXXXXXXXXXX1 = char
|     00000000000000000000000000000001 = 0.0
|
| Pair (i.e. cons cell):
|
|     XXXXXXXXXXXXXXXXXXXXXXXXXXXXX100
|     |     _______________
|     |    |_ _ _ _ _ _ _ _| \ cdr   |
|     |    |_______________| /       | high
|     +--> |_ _ _ _ _ _ _ _| \ car   | memory
|          |_______________| /       V
|           <-- 16 bits -->
|
|
| Hunk (i.e. object in mem. other than pair, e.g. vectors, strings, proc.):
|
|     XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX10
|     |     _______________
|     |    |_ _ _ _ _ _ _ _| field1    |
|     +--> |_______________| field2    | high
|          |_ _ _ _ _ _ _ _| \ dword   | memory
|          |_______________| /   1     V
|          |_ _ _ _ _ _ _ _|     .
|          |_______________|     .
|          |_ _ _ _ _ _ _ _| \ dword
|          |_______________| /   N
|           <-- 16 bits -->
|
|
| Meaning of field1 and field2:
|
| field1            field2               type of object
|
| 000TTTTTNNNNNNNN  NNNNNNNNNNNNNN00     object-vector (T = subtype,
|                                           dword 1..dword N = objects)
| 0010TTTTNNNNNNNN  NNNNNNNNNNNNNN00     mixed-vector (T = subtype,dword N = M,
|                                           dword N-M..dword N = objects)
| 0011TTTTLLLLLLLL  LLLLLLLLLLLLLLLL     byte-vector (T = subtype,
|                                           L = length in bytes)
| 01XXXXXXXXXXXXXX  XXXXXXXXXXXXXXXX     forwarded object (used by GC,
|                                           X = forwarding address / 4)
| 100NNNNNNNNNNNNN  first word of proc.  proc with no closed vars (dword N = M,
|                                           dword N-M..dword N = objects)
| 101XXXXXXXXXXXXX  first word of proc.  predefined procedure
|                                           (these can not be GCed)
| 110NNNNNNNNNNNNN  JSR opcode           closure (dword 1..dword N = closed
|                                           values)
| 111DDDDDDDDDDDDD  first word of code   return point (D = distance in # of
|                                           quad-bytes from parent procedure)
|
| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Garbage collector:
|

| variables:

	AREA(data)

gc_scan_pair:	.long	0
gc_scan_hunk:	.long	0
gc_ret_adr:	.long	0

| for speed, most variables local to the GC are allocated to registers:

space_length	= d0
obj		= d0
space		= d1
temp		= d1
forw_hunk	= d2
dist		= d2
stack_adjust	= d2
count		= d3
forw_pair	= d4
mem_mask	= hunk_mask
heap_start	= d7

adr		= a1
scan		= a2
old		= a6

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	.globl	_garbage_collect

_garbage_collect:

| Initialize

	movl	_libr_ptr,libr_ptr	| get a pointer to the library table

	movl	VAR(heap_ptr),heap_start | get a pointer to the heap and
	movl	VAR(sub_heap_size),space_length | the length of a semi-space

	movl	heap_start,hunk_alloc	| get heap's middle in hunk_alloc
	addl	space_length,hunk_alloc

	cmpl	VAR(hunk_alloc),hunk_alloc | where should we put new space?
	bccs	gc_find_end		| new space is second part of heap

	movl	heap_start,hunk_alloc	| new space is first part

gc_find_end:

	movl	hunk_alloc,pair_alloc	| figure out where new space ends
	addl	space_length,pair_alloc

	movl	pair_alloc,VAR(pair_alloc) | remember what part of the heap
	movl	hunk_alloc,VAR(hunk_alloc) | has been scanned

	moveq	#forwarded_pair,forw_pair | setup registers
	movl	#pair_type_mask+hunk_type_mask,mem_mask
	movl	#length_mask,sp@-	| to get lengths in quad-bytes

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Scan global variables and transfer their values to new space

gc_globals:

	movw	#first_glo+2,scan
	movl	VAR(glob_var_count),count
	lea	gc_globals_return,return_adr
	bras	gc_globals_start

gc_globals_loop:

	movl	global_ptr@(0,scan:w),obj | get object to transfer

	btst	obj,mem_mask	| is the object a memory allocated object?
	beqs	gc_globals_next	| if it isn't, move to next
	cmpl	heap_start,obj	| is the object really in the heap?
	bcc	xfer		| if it is, transfer it to new space

gc_globals_next:

	addqw	#6,scan		| move to next variable

	dbra	count,gc_globals_loop
	bras	gc_globals_exit

gc_globals_return:

	movl	obj,global_ptr@(0,scan:w) | put back transfered object

	addqw	#6,scan		| move to next variable

gc_globals_start:

	dbra	count,gc_globals_loop

gc_globals_exit:

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Scan all hunks and transfer the objects they contain to new space

gc_hunks:

	movl	GLOBAL(hunk_alloc),scan
	lea	gc_hunks_return,return_adr
	bras	gc_hunks_next

gc_hunks_loop:

	movl	scan@+,count	| get header of hunk
	bmis	gc_hunks_proc	| if sign bit is 1, it is a procedure

	movl	count,temp	| compute space allocated for the object
	addql	#3,count	| round to upper multiple of 4 bytes
	andl	sp@,count	| strip type tag and low 2 bits
	bnes	gc_hunks_switch	| if length > 0, the minimum does not apply
	moveq	#4,count	| minimum of 4 bytes must be allocated

gc_hunks_switch:

	movl	scan@,old		| get address of object in old space

	lsll	#3,temp			| what type of vector is it?
	bccs	gc_hunks_copy3_loop	| it's an object-vector
	bmis	gc_hunks_copy1_loop	| it's a byte-vector

					| else, it's a mixed-vector

	cmpb	#stack_type,scan@(-4)	| is it a stack?
	bnes	gc_hunks_mixed		| if it isn't, copy both parts

	movl	scan@(4),stack_adjust	| get the stack's adjustment
	addl	stack_adjust,scan@(-4)	| adjust the header

	addl	count,old		| skip garbage part of stack
	addl	count,scan
	addl	stack_adjust,scan

	movl	old@(-4),count		| get length of active part
	lsrl	#1,count

	subl	count,old		| move to start of active part
	subl	count,scan

	bras	gc_hunks_copy3_loop	| copy active part of stack

gc_hunks_copy1_loop:
	movl	old@+,scan@+	| copy bytes from old space to new space
	subql	#4,count
	bnes	gc_hunks_copy1_loop

	bras	gc_hunks_next	| move to next hunk

gc_hunks_proc:

	movl	scan@,old	| get address of object in old space

	subw	count,count	| put field1 in lower bits of count
	swap	count		| and 0 in higher bits

	lslw	#2,count	| is it a closure?
	bcss	gc_hunks_copy3_loop | if it is, handle it like an object vector

gc_hunks_mixed:

	movl	count,space	| remember total space allocated

	movl	old@(-4,count:l),count | get length of active part
	lsrl	#1,count

	subl	count,space	| compute length of non-active part

gc_hunks_copy2_loop:
	movl	old@+,scan@+	| copy bytes from old space to new space
	subql	#4,space
	bnes	gc_hunks_copy2_loop

gc_hunks_copy3_loop:

	movl	old@+,obj	| get object to transfer from old space

	btst	obj,mem_mask	| is the object a memory allocated object?
	beqs	gc_hunks_return	| if it isn't, move to next
	cmpl	heap_start,obj	| is the object really in the heap?
	bccs	xfer		| if it is, transfer it to new space

gc_hunks_return:

	movl	obj,scan@+	| put transfered object into new space

	subql	#4,count
	bnes	gc_hunks_copy3_loop

gc_hunks_next:

	cmpl	scan,hunk_alloc	| have we scanned all the hunks?
	bnes	gc_hunks_loop

	movl	scan,GLOBAL(hunk_alloc)	| remember where scan ended

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Scan all pairs and transfer their car and cdr to new space

gc_pairs:

	movl	GLOBAL(pair_alloc),scan
	lea	gc_pairs_return,return_adr
	bras	gc_pairs_start

gc_pairs_loop:

	movl	scan@-,obj	| get object to transfer

	btst	obj,mem_mask	| is the object a memory allocated object?
	beqs	gc_pairs_next	| if it isn't, move to next
	cmpl	heap_start,obj	| is the object really in the heap?
	bccs	xfer		| if it is, transfer it to new space

gc_pairs_next:

	cmpl	scan,pair_alloc	| have we scanned all the pairs?
	bnes	gc_pairs_loop
	bras	gc_pairs_exit

gc_pairs_return:

	movl	obj,scan@	| put back transfered object

gc_pairs_start:

	cmpl	scan,pair_alloc	| have we scanned all the pairs?
	bnes	gc_pairs_loop

gc_pairs_exit:

	movl	scan,GLOBAL(pair_alloc)	| remember where scan ended

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	cmpl	gc_scan_pair,pair_alloc | have any pairs been added?
	bne	gc_pairs	| if so, scan the pairs and hunks again

	movl	pair_alloc,temp		| compute free space count
	subl	hunk_alloc,temp
	subl	GLOBAL(heap_buf),temp	| reserve buffer space
	asrl	#2,temp
	movl	temp,GLOBAL(free_count)

	addql	#4,sp			| deallocate length mask

	rts

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|
| Transfer an object to new space

| we assume that xfer has been called by code similar to:
|
|	lea	ret,return_adr	| setup return address
|	movl	...,obj		| get object to transfer
|	btst	obj,mem_mask	| is the object a memory allocated object?
|	beqs	ret		| if it isn't, do nothing
|	cmpl	heap_start,obj	| is the object really in the heap?
|	bccs	xfer		| if it is, transfer it
|ret:

xfer:				| transfer obj to new space

	movl	obj,adr		| put address of object in an address register

	btst	obj,pair_mask	| is it a pair?
	beqs	xfer_hunk	| if it is not, it is a hunk

xfer_pair:

|     adr
|     |     _______________
|     |    |_ _ _ _ _ _ _ _| \ cdr   |
|     |    |_______________| /       | high
|     +--> |_ _ _ _ _ _ _ _| \ car   | memory
|          |_______________| /       V
|           <-- 16 bits -->

	movl	adr@,temp	| get the car
	cmpl	forw_pair,temp	| has the pair been forwarded?
	beqs	xfer_forw_pair	| if it has, return forwarding address

	movl	forw_pair,adr@	  | leave forwarding indicator in car of pair
	movl	temp,pair_alloc@- | copy car to new space
	movl	adr@-,temp	  | get the cdr
	movl	pair_alloc,adr@	  | leave forwarding address in cdr of pair
	movl	pair_alloc,obj	  | return the forwarding address
	movl	temp,pair_alloc@- | copy cdr to new space

	jmp	return_adr@	| return to caller

xfer_forw_pair:

	movl	adr@-,obj	| return the forwarding address (ie. the cdr)

	jmp	return_adr@	| return to caller

xfer_forw_hunk:

	lsll	#1,obj		| prepare forwarding address
	subql	#2,obj		| adjust forw. addr. (so it ends with 10)

	jmp	return_adr@	| return to caller

xfer_hunk:

|     adr
|     |     _______________
|     |    |_ _ _ _ _ _ _ _| field1    |
|     +--> |_______________| field2    | high
|          |_ _ _ _ _ _ _ _| \ dword   | memory
|          |_______________| /   1     V
|          |_ _ _ _ _ _ _ _|     .
|          |_______________|     .
|          |_ _ _ _ _ _ _ _| \ dword
|          |_______________| /   N
|           <-- 16 bits -->

	subql	#2,adr		| move to start of hunk
	movl	adr@,obj	| get header of hunk
	bmis	xfer_proc	| if sign bit is 1, it is a procedure

| header                               type of object
|
| 00XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX     object-vect., mixed-vect. or byte-vect.
| 01XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX     forwarded object

	lsll	#1,obj		| test forwarded bit
	bmis	xfer_forw_hunk	| return forw. addr. if it has been forwarded

	lsrl	#1,obj		| restore old header
	movl	obj,hunk_alloc@+ | and copy it to new space

	movl	obj,space	| compute space to allocate for the object
	addql	#3,space	| round to upper multiple of 4 bytes
	andl	sp@,space	| strip type tag and low 2 bits
	bnes	xfer_stack	| if length > 0, the minimum does not apply
	moveq	#4,space	| minimum of 4 bytes must be allocated

xfer_stack:

	cmpb	#stack_type,adr@ | is it a stack?
	bnes	xfer_alloc	| if it isn't simply allocate the space needed

| stacks should be shrunk if their active part is small compared to their
| total size (to save space) or they should be extended if their active part
| is close to their total size (to prevent GCs caused by stack overflows).

	movl	adr@(0,space:l),stack_adjust | get length of active part
	lsrl	#1,stack_adjust		| convert from scheme fixnum
	addl	GLOBAL(stack_free),stack_adjust
	addl	GLOBAL(stack_buf),stack_adjust
	subl	space,stack_adjust

	movl	stack_adjust,hunk_alloc@(4) | remember the adjustment
	addl	stack_adjust,space | change size

xfer_alloc:

	movl	hunk_alloc,obj	| return the address of the object in new space
	subql	#2,obj

	movl	hunk_alloc,forw_hunk | prepare forwarding address
	addqw	#1,forw_hunk	| set forwarded bit
	rorl	#2,forw_hunk	| setup header format for forwarded object
	movl	forw_hunk,adr@+	| change header to forwarding address

	movl	adr,hunk_alloc@	| keep a pointer to the old object in the copy

	addl	space,hunk_alloc | allocate space needed for object

	jmp	return_adr@	| return to caller

xfer_proc:

	movl	obj,space
	swap	space		| put field1 in lower bits of space

| field1            type of object
|
| 100NNNNNNNNNNNNN  procedure with no closed vars (N = length in quad-bytes)
| 110NNNNNNNNNNNNN  closure (N = length in quad-bytes)
| 111DDDDDDDDDDDDD  return point (D = dist. in quad-bytes from parent proc.)
|
| note: predefined procedures (ie. field1 = 101XXXXXXXXXXXXX) do not have to
|       be handled because they are not really on the heap

	lslw	#2,space
	bmis	xfer_ret_point

	movl	obj,hunk_alloc@+ | copy header to new space

	movl	hunk_alloc,obj	| return the address of the object in new space
	subql	#2,obj

	movl	hunk_alloc,forw_hunk | prepare forwarding address
	addqw	#1,forw_hunk	| set forwarded bit
	rorl	#2,forw_hunk	| setup header format for forwarded object
	movl	forw_hunk,adr@+	| change header to forwarding address

	movl	adr,hunk_alloc@	| keep a pointer to the old object in the copy

	addw	space,hunk_alloc | allocate space needed for object

	jmp	return_adr@	| return to caller

xfer_ret_point:

	extl	space		| turn distance to a 32 bit value

	addl	space,adr	| move to parent procedure
	subql	#2,adr		| move to start of procedure
	movl	adr@,obj	| get header of procedure
	bmis	xfer_parent	| if sign bit is 1, it hasn't been forwarded

	lsll	#2,obj		| extract forwarding address
	subql	#2,obj		| adjust forw. addr. (so it ends with 10)
	subl	space,obj	| move back to return point

	jmp	return_adr@	| return to caller

xfer_parent:

	movl	space,dist
	extl	dist		| turn distance to a 32 bit value

	movl	obj,space
	swap	space		| put field1 in lower bits of space
	lslw	#2,space

	movl	obj,hunk_alloc@+ | copy header to new space

	movl	hunk_alloc,obj	| get the address of the parent in new space
	subql	#2,obj
	subl	dist,obj	| move back to return point

	movl	hunk_alloc,forw_hunk | prepare forwarding address
	addqw	#1,forw_hunk	| set forwarded bit
	rorl	#2,forw_hunk	| setup header format for forwarded object
	movl	forw_hunk,adr@+	| change header to forwarding address

	movl	adr,hunk_alloc@	| keep a pointer to the old object in the copy

	addw	space,hunk_alloc | allocate space needed for object

	jmp	return_adr@	| return to caller

| - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
