;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; File:         deriv.sch
; Description:  The DERIV benchmark from the Gabriel tests.
; Author:       Vaughan Pratt
; Created:      8-Apr-85
; Modified:     10-Apr-85 14:53:50 (Bob Shaw)
;               23-Jul-87 (Will Clinger)
;               9-Feb-88 (Will Clinger)
; Language:     Scheme
; Status:       Public Domain
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
;;; DERIV -- Symbolic derivative benchmark written by Vaughn Pratt.
;;; It uses a simple subset of Lisp and does a lot of  CONSing.

; Returns the wrong answer for quotients.
; Fortunately these aren't used in the benchmark.

(define (run)
 
  (define (deriv-aux a) (list '/ (deriv a) a))
 
  (define (deriv a)
    (cond
      ((not (pair? a))
       (cond ((eq? a 'x) 1) (else 0)))
      ((eq? (car a) '+)
       (cons '+ (map deriv (cdr a))))
      ((eq? (car a) '-)
       (cons '- (map deriv
                        (cdr a))))
      ((eq? (car a) '*)
       (list '*
             a
             (cons '+ (map deriv-aux (cdr a)))))
      ((eq? (car a) '/)
       (list '-
             (list '/
                   (deriv (cadr a))
                   (caddr a))
             (list '/
                   (cadr a)
                   (list '*
                         (caddr a)
                         (caddr a)
                         (deriv (caddr a))))))
      (else 'error)))
 
  (do ((i 0 (+ i 1)))
      ((= i 1000))
    (deriv '(+ (* 3 x x) (* a x x) (* b x) 5))
    (deriv '(+ (* 3 x x) (* a x x) (* b x) 5))
    (deriv '(+ (* 3 x x) (* a x x) (* b x) 5))
    (deriv '(+ (* 3 x x) (* a x x) (* b x) 5))
    (deriv '(+ (* 3 x x) (* a x x) (* b x) 5))))
 
;;; call:  (run)

(define (run-benchmark name thunk)
  (display name) (display " benchmark") (newline)
  (newline)
  (let loop ((thunk thunk) (start (cpu-time)) (i 1))
    (let ((result (thunk)))
      (let ((delta (- (cpu-time) start)))
        (if (< delta 600) ; run for at least 10 seconds
          (loop thunk start (+ i 1))
          (let ((time-per-call (quotient (quotient (* delta 100) 60) i)))
            (display "called test procedure ") (display i)
            (display " times in ") (display delta)
            (display "/60 seconds of user cpu time.") (newline)
            (display "average user cpu time per call: ")
            (let ((seconds (quotient time-per-call 100))
                  (hundreds (remainder time-per-call 100)))
              (display seconds) (display (if (< hundreds 10) ".0" "."))
              (display hundreds)
              (newline)
              (newline)
              result)))))))

(run-benchmark "Deriv" (lambda () (run)))
