;_______________________________________________________________________________________________
;
;                             Interpreteur Scheme ecrit en Scheme
;
; Note: Seulement les formes speciales 'quote', 'set!', 'if' et 'lambda' ainsi que la reference
;       de constante et de variable et l'application de fonction sont traitees car les autres
;       constructions de Scheme peuvent s'exprimer a partir de celles-ci.
;_______________________________________________________________________________________________


(define (interpret expr)
  (int expr *env-glo*))

(define (int expr env)
  (cond ((symbol? expr)          (int-ref expr env))
        ((not (pair? expr))      (int-cst expr env))
        ((eq? (car expr) 'quote) (int-cst (cadr expr) env))
        ((eq? (car expr) 'set!)  (int-set (cadr expr) (caddr expr) env))
        ((eq? (car expr) 'if)    (int-tst (cadr expr) (caddr expr) (cadddr expr) env))
        ((eq? (car expr) 'lambda)
         (let ((p (cadr expr)))
           (cond ((null? p)          (int-fn0      (caddr expr) env))
                 ((symbol? p)        (int-fn1/rest (caddr expr) p env))
                 ((null? (cdr p))    (int-fn1      (caddr expr) (car p) env))
                 ((symbol? (cdr p))  (int-fn2/rest (caddr expr) (car p) (cdr p) env))
                 ((null? (cddr p))   (int-fn2      (caddr expr) (car p) (cadr p) env))
                 ((symbol? (cddr p)) (int-fn3/rest (caddr expr) (car p) (cadr p) (cddr p) env))
                 ((null? (cdddr p))  (int-fn3      (caddr expr) (car p) (cadr p) (caddr p) env))
                 (else               (error "trop de parametres dans une lambda-expression")))))
        ((null? (cdr expr))    (int-ap0 (car expr) env))
        ((null? (cddr expr))   (int-ap1 (car expr) (cadr expr) env))
        ((null? (cdddr expr))  (int-ap2 (car expr) (cadr expr) (caddr expr) env))
        ((null? (cddddr expr)) (int-ap3 (car expr) (cadr expr) (caddr expr) (cadddr expr) env))
        (else                  (error "trop d'arguments dans une application de fonction"))))

; --- interpretation d'une constante ---

(define (int-cst a env)
  a)

; --- interpretation d'une reference de variable ---

(define (int-ref a env)
  (cdr (assq a env)))

; --- interpretation d'une affectation ---

(define (int-set a b env)
  (set-cdr! (assq a env) (int b env)))

; --- interpretation d'une forme speciale 'if' ---

(define (int-tst a b c env)
  (if (int a env) (int b env) (int c env)))

; --- interpretation d'une application de fonction ---

(define (int-ap0 a env)
  ((int a env)))

(define (int-ap1 a b env)
  ((int a env) (int b env)))

(define (int-ap2 a b c env)
  ((int a env) (int b env) (int c env)))

(define (int-ap3 a b c d env)
  ((int a env) (int b env) (int c env) (int d env)))

; --- interpretation d'une forme speciale 'lambda' ---

(define (int-fn0 a env)
  (lambda () (int a env)))

(define (int-fn1 a b env)
  (lambda (x) (int a (cons (cons b x) env))))

(define (int-fn2 a b c env)
  (lambda (x y) (int a (cons (cons b x) (cons (cons c y) env)))))

(define (int-fn3 a b c d env)
  (lambda (x y z) (int a (cons (cons b x) (cons (cons c y) (cons (cons d z) env))))))

(define (int-fn1/rest a b env)
  (lambda x (int a (cons (cons b x) env))))

(define (int-fn2/rest a b c env)
  (lambda (x . y) (int a (cons (cons b x) (cons (cons c y) env)))))

(define (int-fn3/rest a b c d env)
  (lambda (x y . z) (int a (cons (cons b x) (cons (cons c y) (cons (cons d z) env))))))

; --- definition de variables globales ---

(define (define-global var val)
  (if (assq var *env-glo*)
    (set-cdr! (assq var *env-glo*) val)
    (begin
      (set-cdr! *env-glo* (cons (car *env-glo*) (cdr *env-glo*)))
      (set-car! *env-glo* (cons var val)))))

(define *env-glo* (list (cons 'define define-global)))
(define-global 'cons  cons )
(define-global 'car   car  )
(define-global 'cdr   cdr  )
(define-global 'null? null?)
(define-global 'not   not  )
(define-global '<     <    )
(define-global '+     +    )
(define-global '-     -    )

; --- pour evaluer, on appelle l'interprete ---

(define (evaluer expr)
  (interpret expr))


;_______________________________________________________________________________________________
;
; Exemples d'utilisation de l'interprete:
;
; a) pour evaluer:
;
;      (define fib
;        (lambda (x)
;          (if (< x 2) x (+ (fib (- x 1)) (fib (- x 2))))))
;
;    il faut entrer:
;
       (evaluer '(define 'fib
                   (lambda (x)
                     (if (< x 2) x (+ (fib (- x 1)) (fib (- x 2)))))))
;
; b) pour evaluer:
;
;      (fib 20)
;
;    il faut entrer:
;
       (evaluer '(fib 20))
