;------------------------------------------------------------------------------

(define (queue-empty)         ; the empty set
  (cons '() '()))

(define (queue-empty? queue)  ; is the queue empty?
  (null? (car queue)))

(define (queue-get! queue)    ; remove the first element of the queue
  (if (null? (car queue))
    (error "Can not remove an element from an empty queue")
    (let ((x (caar queue)))
      (set-car! queue (cdar queue))
      (if (null? (car queue)) (set-cdr! queue '()))
      x)))

(define (queue-put! queue x)  ; add an element to the end of the queue
  (let ((entry (cons x '())))
    (if (null? (car queue))
      (set-car! queue entry)
      (set-cdr! (cdr queue) entry))
    (set-cdr! queue entry)
    x))

;------------------------------------------------------------------------------

(define task-queue '())

(define (task-start proc)
  (queue-put! task-queue (current-process))
  (call-with-current-continuation (lambda (cont) (cont '()) (proc))))
  
(define (task-switch)
  (queue-put! task-queue (current-process))
  ((queue-get! task-queue) '()))

(define (task-terminate)
  (if (queue-empty? task-queue)
    (error "All tasks have terminated")
    ((queue-get! task-queue) '())))

;------------------------------------------------------------------------------

(define (test)

  (define (process1)
    (do ((i 1 (+ i 1))) ((= i 10))
      (display "process1 ") (display i) (newline)
      (task-switch))
    (task-terminate))

  (define (process2)
    (do ((i 1 (+ i 1))) ((= i 5))
      (display "process2 ") (display i) (newline)
      (task-switch))
    (task-terminate))

  (define (process3)
    (do ((i 1 (+ i 1))) ((= i 8))
      (display "process3 ") (display i) (newline)
      (task-switch))
    (task-terminate))

  (define (process4)
    (do ((i 1 (+ i 1))) ((= i 9))
      (display "process4 ") (display i) (newline)
      (task-switch))
    (task-terminate))

  (set! task-queue (queue-empty))

  (task-start process1)
  (task-start process2)
  (task-start process3)
  (task-start process4)

  (task-terminate))

(test)
