(define (interpret expr)
  (int expr *env-glo*))

(define (int expr env)
  (cond ((symbol? expr)
         (cdr (assq expr env)))
        ((not (pair? expr))
         expr)
        ((eq? (car expr) 'quote)
         (cadr expr))
        ((eq? (car expr) 'set!)
         (set-cdr! (assq (cadr expr) env) (int (caddr expr) env)))
        ((eq? (car expr) 'if)
         (if (int (cadr expr) env) (int (caddr expr) env) (int (cadddr expr) env)))
        ((eq? (car expr) 'define)
         (if (pair? (cadr expr))
           (define-global (car (cadr expr))
                          (int (list 'lambda (cdr (cadr expr)) (caddr expr)) env))
           (define-global (cadr expr) (int (caddr expr) env))))
        ((eq? (car expr) 'lambda)
         (let ((p (cadr expr)) (b (caddr expr)))
           (lambda l (int b (alloc p l env)))))
        (else
         (apply (int (car expr) env) (int-args (cdr expr) env)))))

(define (alloc p l env)
  (cond ((null? p)   env)
        ((symbol? p) (cons (cons p l) env))
        (else        (cons (cons (car p) (car l)) (alloc (cdr p) (cdr l) env)))))

(define (int-args l env)
  (if (null? l) '() (cons (int (car l) env) (int-args (cdr l) env))))

; --- definition de variables globales ---

(define (define-global var val)
  (if (assq var *env-glo*)
    (set-cdr! (assq var *env-glo*) val)
    (begin
      (set-cdr! *env-glo* (cons (car *env-glo*) (cdr *env-glo*)))
      (set-car! *env-glo* (cons var val))))
  var)

(define *env-glo* (list (cons 'not not)))
(define-global 'BOOLEAN? BOOLEAN?)
(define-global 'EQV? EQV?)
(define-global 'EQ? EQ?)
(define-global 'EQUAL? EQUAL?)
(define-global 'PAIR? PAIR?)
(define-global 'CONS CONS)
(define-global 'CAR CAR)
(define-global 'CDR CDR)
(define-global 'SET-CAR! SET-CAR!)
(define-global 'SET-CDR! SET-CDR!)
(define-global 'CAAR CAAR)
(define-global 'CADR CADR)
(define-global 'CDAR CDAR)
(define-global 'CDDR CDDR)
(define-global 'CAAAR CAAAR)
(define-global 'CAADR CAADR)
(define-global 'CADAR CADAR)
(define-global 'CADDR CADDR)
(define-global 'CDAAR CDAAR)
(define-global 'CDADR CDADR)
(define-global 'CDDAR CDDAR)
(define-global 'CDDDR CDDDR)
(define-global 'CAAAAR CAAAAR)
(define-global 'CAAADR CAAADR)
(define-global 'CAADAR CAADAR)
(define-global 'CAADDR CAADDR)
(define-global 'CADAAR CADAAR)
(define-global 'CADADR CADADR)
(define-global 'CADDAR CADDAR)
(define-global 'CADDDR CADDDR)
(define-global 'CDAAAR CDAAAR)
(define-global 'CDAADR CDAADR)
(define-global 'CDADAR CDADAR)
(define-global 'CDADDR CDADDR)
(define-global 'CDDAAR CDDAAR)
(define-global 'CDDADR CDDADR)
(define-global 'CDDDAR CDDDAR)
(define-global 'CDDDDR CDDDDR)
(define-global 'NULL? NULL?)
(define-global 'LIST LIST)
(define-global 'LENGTH LENGTH)
(define-global 'APPEND APPEND)
(define-global 'REVERSE REVERSE)
(define-global 'LIST-TAIL LIST-TAIL)
(define-global 'LIST-REF LIST-REF)
(define-global 'LAST-PAIR LAST-PAIR)
(define-global 'MEMQ MEMQ)
(define-global 'MEMV MEMV)
(define-global 'MEMBER MEMBER)
(define-global 'ASSQ ASSQ)
(define-global 'ASSV ASSV)
(define-global 'ASSOC ASSOC)
(define-global 'SYMBOL? SYMBOL?)
(define-global 'SYMBOL->STRING SYMBOL->STRING)
(define-global 'STRING->SYMBOL STRING->SYMBOL)
(define-global 'NUMBER? NUMBER?)
(define-global 'REAL? REAL?)
(define-global 'INTEGER? INTEGER?)
(define-global 'ODD? ODD?)
(define-global 'EVEN? EVEN?)
(define-global '= =)
(define-global '< <)
(define-global '> >)
(define-global '<= <=)
(define-global '>= >=)
(define-global 'MAX MAX)
(define-global 'MIN MIN)
(define-global '+ +)
(define-global '* *)
(define-global '- -)
(define-global '/ /)
(define-global 'ABS ABS)
(define-global 'QUOTIENT QUOTIENT)
(define-global 'REMAINDER REMAINDER)
(define-global 'EXPT EXPT)
(define-global 'NUMBER->STRING NUMBER->STRING)
(define-global 'STRING->NUMBER STRING->NUMBER)
(define-global 'CHAR? CHAR?)
(define-global 'CHAR=? CHAR=?)
(define-global 'CHAR<? CHAR<?)
(define-global 'CHAR>? CHAR>?)
(define-global 'CHAR<=? CHAR<=?)
(define-global 'CHAR>=? CHAR>=?)
(define-global 'CHAR->INTEGER CHAR->INTEGER)
(define-global 'INTEGER->CHAR INTEGER->CHAR)
(define-global 'CHAR-UPPERCASE CHAR-UPPERCASE)
(define-global 'STRING? STRING?)
(define-global 'MAKE-STRING MAKE-STRING)
(define-global 'STRING-LENGTH STRING-LENGTH)
(define-global 'STRING-REF STRING-REF)
(define-global 'STRING-SET! STRING-SET!)
(define-global 'STRING=? STRING=?)
(define-global 'STRING->LIST STRING->LIST)
(define-global 'LIST->STRING LIST->STRING)
(define-global 'VECTOR? VECTOR?)
(define-global 'MAKE-VECTOR MAKE-VECTOR)
(define-global 'VECTOR VECTOR)
(define-global 'VECTOR-LENGTH VECTOR-LENGTH)
(define-global 'VECTOR-REF VECTOR-REF)
(define-global 'VECTOR-SET! VECTOR-SET!)
(define-global 'VECTOR->LIST VECTOR->LIST)
(define-global 'LIST->VECTOR LIST->VECTOR)
(define-global 'PROCEDURE? PROCEDURE?)
(define-global 'PROCEDURE-INFO PROCEDURE-INFO)
(define-global 'APPLY APPLY)
(define-global 'MAP MAP)
(define-global 'FOR-EACH FOR-EACH)
(define-global 'MAKE-DELAYED MAKE-DELAYED)
(define-global 'FORCE FORCE)
(define-global 'CALL-WITH-CURRENT-CONTINUATION CALL-WITH-CURRENT-CONTINUATION)
(define-global 'READ READ)
(define-global 'EOF-OBJECT? EOF-OBJECT?)
(define-global 'READ-CHAR READ-CHAR)
(define-global 'PEEK-CHAR PEEK-CHAR)
(define-global 'WRITE* WRITE*)
(define-global 'WRITE WRITE)
(define-global 'DISPLAY DISPLAY)
(define-global 'NEWLINE NEWLINE)
(define-global 'WRITE-CHAR WRITE-CHAR)
(define-global 'ERROR ERROR)

; --- pour evaluer, on appelle l'interprete ---

(define (eval expr)
  (interpret expr))

(define (top-level)
  (display "> ")
  (let ((x (read)))
    (if (not (eof-object? x))
      (begin
        (write (eval x)) (newline)
        (newline)
        (top-level)))))

(top-level)
